@@ -45,6 +45,7 @@ struct connman_service {
 	unsigned int order;
 	char *name;
 	char *passphrase;
+	char *profile;
 	struct connman_device *device;
 	struct connman_network *network;
 };
@@ -289,27 +290,20 @@ static DBusMessage *connect_service(DBusConnection *conn,
 	if (service->network != NULL) {
 		int err;
 
+		connman_network_set_string(service->network,
+				"WiFi.Passphrase", service->passphrase);
+
 		err = connman_network_connect(service->network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
-
-		service->state = CONNMAN_SERVICE_STATE_ASSOCIATION;
-
-		state_changed(service);
+			return __connman_error_failed(msg, -err);
 
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
-
-	if (service->device != NULL) {
+	} else if (service->device != NULL) {
 		if (service->favorite == FALSE)
 			return __connman_error_no_carrier(msg);
 
 		if (__connman_device_connect(service->device) < 0)
-			return __connman_error_failed(msg);
-
-		service->state = CONNMAN_SERVICE_STATE_READY;
-
-		state_changed(service);
+			return __connman_error_failed(msg, EINVAL);
 
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 	}
@@ -327,25 +321,18 @@ static DBusMessage *disconnect_service(DBusConnection *conn,
 
 		err = __connman_network_disconnect(service->network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
-
-		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
-
-		state_changed(service);
+			return __connman_error_failed(msg, -err);
 
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
-	}
+	} else if (service->device != NULL) {
+		int err;
 
-	if (service->device != NULL) {
 		if (service->favorite == FALSE)
 			return __connman_error_no_carrier(msg);
 
-		if (__connman_device_connect(service->device) < 0)
-			return __connman_error_failed(msg);
-
-		service->state = CONNMAN_SERVICE_STATE_IDLE;
-
-		state_changed(service);
+		err = __connman_device_disconnect(service->device);
+		if (err < 0)
+			return __connman_error_failed(msg, -err);
 
 		return g_dbus_create_reply(msg, DBUS_TYPE_INVALID);
 	}
@@ -366,11 +353,7 @@ static DBusMessage *remove_service(DBusConnection *conn,
 
 		err = __connman_network_disconnect(service->network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
-
-		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
-
-		state_changed(service);
+			return __connman_error_failed(msg, -err);
 	}
 
 	connman_service_set_favorite(service, FALSE);
@@ -440,6 +423,7 @@ static void service_free(gpointer data)
 	if (service->network != NULL)
 		connman_network_unref(service->network);
 
+	g_free(service->profile);
 	g_free(service->name);
 	g_free(service->passphrase);
 	g_free(service->identifier);
@@ -595,54 +579,49 @@ int __connman_service_set_carrier(struct connman_service *service,
 		break;
 	}
 
-	if (carrier == TRUE)
-		service->state = CONNMAN_SERVICE_STATE_CARRIER;
-	else
-		service->state = CONNMAN_SERVICE_STATE_IDLE;
+	if (carrier == FALSE) {
+		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
+		state_changed(service);
 
-	state_changed(service);
+		service->state = CONNMAN_SERVICE_STATE_IDLE;
+		state_changed(service);
+	} else {
+		service->state = CONNMAN_SERVICE_STATE_CARRIER;
+		state_changed(service);
+	}
 
 	return connman_service_set_favorite(service, carrier);
 }
 
-int __connman_service_indicate_configuration(struct connman_service *service)
+int __connman_service_indicate_state(struct connman_service *service,
+					enum connman_service_state state)
 {
-	DBG("service %p", service);
+	DBG("service %p state %d", service, state);
 
 	if (service == NULL)
 		return -EINVAL;
 
-	service->state = CONNMAN_SERVICE_STATE_CONFIGURATION;
+	if (state == CONNMAN_SERVICE_STATE_CARRIER)
+		return __connman_service_set_carrier(service, TRUE);
 
-	state_changed(service);
+	if (service->state == state)
+		return -EALREADY;
 
-	return 0;
-}
-
-int __connman_service_ready(struct connman_service *service)
-{
-	DBG("service %p", service);
-
-	if (service == NULL)
+	if (service->state == CONNMAN_SERVICE_STATE_IDLE &&
+				state == CONNMAN_SERVICE_STATE_DISCONNECT)
 		return -EINVAL;
 
-	service->state = CONNMAN_SERVICE_STATE_READY;
+	if (state == CONNMAN_SERVICE_STATE_IDLE &&
+			service->state != CONNMAN_SERVICE_STATE_DISCONNECT) {
+		service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
+		state_changed(service);
+	}
 
+	service->state = state;
 	state_changed(service);
 
-	return 0;
-}
-
-int __connman_service_disconnect(struct connman_service *service)
-{
-	DBG("service %p", service);
-
-	if (service == NULL)
-		return -EINVAL;
-
-	service->state = CONNMAN_SERVICE_STATE_DISCONNECT;
-
-	state_changed(service);
+	if (state == CONNMAN_SERVICE_STATE_READY)
+		return connman_service_set_favorite(service, TRUE);
 
 	return 0;
 }
@@ -693,6 +672,8 @@ struct connman_service *connman_service_get(const char *identifier)
 
 	service->identifier = g_strdup(identifier);
 
+	service->profile = g_strdup(__connman_profile_active_ident());
+
 	__connman_storage_load_service(service);
 
 	iter = g_sequence_insert_sorted(service_list, service,
@@ -986,15 +967,94 @@ done:
 
 static int service_load(struct connman_service *service)
 {
+	GKeyFile *keyfile;
+	gchar *pathname, *data = NULL;
+	gsize length;
+	char *str;
+
 	DBG("service %p", service);
 
+	if (service->profile == NULL)
+		return -EINVAL;
+
+	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	if (pathname == NULL)
+		return -ENOMEM;
+
+	keyfile = g_key_file_new();
+
+	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE) {
+		g_free(pathname);
+		return -ENOENT;
+	}
+
+	g_free(pathname);
+
+	if (g_key_file_load_from_data(keyfile, data, length,
+							0, NULL) == FALSE) {
+		g_free(data);
+		return -EILSEQ;
+	}
+
+	g_free(data);
+
+	str = g_key_file_get_string(keyfile,
+				service->identifier, "Passphrase", NULL);
+	if (str != NULL) {
+		g_free(service->passphrase);
+		service->passphrase = str;
+	}
+
+	g_key_file_free(keyfile);
+
 	return 0;
 }
 
 static int service_save(struct connman_service *service)
 {
+	GKeyFile *keyfile;
+	gchar *pathname, *data = NULL;
+	gsize length;
+
 	DBG("service %p", service);
 
+	if (service->profile == NULL)
+		return -EINVAL;
+
+	pathname = g_strdup_printf("%s/%s.conf", STORAGEDIR, service->profile);
+	if (pathname == NULL)
+		return -ENOMEM;
+
+	keyfile = g_key_file_new();
+
+	if (g_file_get_contents(pathname, &data, &length, NULL) == FALSE)
+		goto update;
+
+	if (length > 0) {
+		if (g_key_file_load_from_data(keyfile, data, length,
+							0, NULL) == FALSE)
+			goto done;
+	}
+
+	g_free(data);
+
+update:
+	if (service->passphrase != NULL)
+		g_key_file_set_string(keyfile, service->identifier,
+					"Passphrase", service->passphrase);
+
+	data = g_key_file_to_data(keyfile, &length, NULL);
+
+	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
+		connman_error("Failed to store service information");
+
+done:
+	g_free(data);
+
+	g_key_file_free(keyfile);
+
+	g_free(pathname);
+
 	return 0;
 }
 
